 /***********************************************************
  *
  *   Program:    Routers
  *   Created:    01/27/01 -- 02/11/01
  *   Author:     V.Molotkov
  *   Comments:   Third test in 2001 Programmer's contest
	*								of PC Magazine(Bg)
  *
  ************************************************************
  *
  * History:
  *
  * $Log:$
  *
  ************************************************************/

#include <stdio.h>
#include <stdlib.h>
#pragma warning(disable: 4786)

#include <vector>
#include <list>
#include <cassert>
#include <iostream>
#include <fstream>
#include <limits>
#include <iomanip>
#include "timer.h"

using namespace std;

#define FIND_SHORTEST_ROUTE //Not necessarily optimal
#ifdef FIND_SHORTEST_ROUTE
#define OPTIMIZE_EQUAL_LINKS//If (Q,P) are almost equal for all links,
														//the shortest path is the optimal one;
#endif
#define USE_VECTOR_ROUTES//Route as vector instead of list
#define CHOOSE_SHORTER_BEST //Not necessarily of minimal length
#define OPTIMIZE_SET_DIST//This turns sometimes SetDistTo to O(N) instead of O(N*N*N)
//To be defined always (0 == switch off, max==3):
#define FIND_BAD_LINKS 2//If>0 this turns O(exp(N)) (in general case)
											 //algorithm into O(N*N*N) in case when
											 //(almost) all P's are equal to 1.0;
											 //if > 1, bad links are searched in NextFirst()
											 //in addition to that of Next(),
											 //so the routine finds bad links at earlier stages
											 //if == 3 speeds up greatly search in polynomial case,
											 //but there are doubts yet in its correctness
#if FIND_BAD_LINKS>0
#define CUTOFF
#endif
#if FIND_BAD_LINKS>1
#define QUICK_MOUNT//Speed up to 1.5-3 times approx. (for the case with all P near to 1)
#define CACHE_WOULDBE_BAD_LINKS//Potentially bad links are cached in badInd[]
#if undef FIND_SHORTEST_ROUTE
#define FIND_SHORTEST_ROUTE//Otherwise First() may fail;
#endif
#endif

#ifdef QUICK_MOUNT
#if undef USE_VECTOR_ROUTES
#define USE_VECTOR_ROUTES//Route as vector instead of list
#endif
#endif

#ifdef CUTOFF
#if undef OPTIMIZE_SET_DIST
#define OPTIMIZE_SET_DIST
#endif
#endif

#define APPROX_BEST//bestRoute search is within the given precision epsilon for effectivity

#define SET_BAD_LINKS//Labels bad links, comparing link effectivity with currBest

//Debug Flags:
//#define SHOW_DIST
//#define SHOW_BAD_LINKS
#define STATS

const int MAX_DIST = 2048;//"infinite" distance between two routers
#ifdef APPROX_BEST
//const// better make it variable: changing precision may speed up search
double epsilon = 0.0000005;//Default precision value for effectivity
#endif

inline int min(int i,int j){return i<=j?i:j;}
inline double min(double i,double j){return i<=j?i:j;}
inline int max(int i,int j){return i<=j?j:i;}
inline double max(double i,double j){return i<=j?j:i;}
inline double dist(double x,double y){return x<=y?y-x:x-y;}
inline bool equiv(double x,double y){return(x<=y?y-x:x-y)<epsilon;}

#ifdef APPROX_BEST
#define EQUIV(x,y) equiv(x,y)
#define LEQ(x,y) x<y||equiv(x,y)
#define LESS(x,y) y-x>epsilon
#else
#define EQUIV(x,y) x==y
#define LEQ(x,y) x<=y
#define LESS(x,y) x<y
#endif

//**********************************************************
struct Channel{
	int from;//source router
	int to;//target router
	int Q;//speed
	double P;//safety
	int indF,indT;//second index in the double array. Saved for quick access to the next channel
	bool isBad;//Some channels will be labeled bad in the process of iteration
	Channel(int f=0,int t=0,int q=0,double p=0.0,int iF=0,int iT=0):
		from(f),to(t),P(p),Q(q),indF(iF),indT(iT),isBad(false){}
};

//**********************************************************
//Save "integral" Q and P for every link
//of the route to avoid their recalculation
//(iterating over the head subroute)
struct ChannelX{
	Channel* channel;
	int Q;
	double P;
	ChannelX(Channel* cp=0,int q=0,double p=0):channel(cp),Q(q),P(p){}
};

//**********************************************************

#ifndef USE_VECTOR_ROUTES
typedef list<ChannelX*> Route;
#else
typedef vector<ChannelX*> Route;//Seems to be a bit better
#endif

inline void append(Route& r, Channel* pc){
	int q=pc->Q;
	double p=pc->P;
	if(r.size()!=0){
		Route::reverse_iterator it = r.rbegin();
		p *= (*it)->P;
		q = min(q,(*it)->Q);
	}
	ChannelX* pcX = new ChannelX(pc,q,p);//Seems to be the reason for memory leaks!
	r.push_back(pcX);
}

#ifndef BUG3//memory leaks, eating heap
//to be used instead of r.pop_back(); for non-empty r only!
inline void popAndClean(Route& r){
	ChannelX* pc = *r.rbegin();
	r.pop_back();
	delete pc;
}

//The above stops memory leaks but creates another problem:
//operator=() for Route becomes inappropriate for copyng currRoute
//to bestRoute: it copies only pointers to ChannelX; but the latter
//become inadequate, if deleted. So we are to create a more appropriate
//Route copying routine:
inline void clear(Route&r){
	Route::iterator it = r.begin();
	while(it!=r.end()){
		delete *it;
		++it;
	}
	r.clear();
}

inline void copy(Route&targ,const Route&src){
	clear(targ);
	Route::const_iterator it = src.begin();
	while(it!=src.end()){
		ChannelX* chX = new ChannelX((*it)->channel,(*it)->Q,(*it)->P);
		targ.push_back(chX);
		++it;
	}
}
#endif

inline double GetE(const Route& r){
	Route::const_reverse_iterator it = r.rbegin();
	return it!=r.rend()?(double)((*it)->Q*(*it)->P):-1;
}

#ifndef BUG2
inline void GetQPprev(const Route& r,int& Q,double&P){
	Route::const_reverse_iterator it = r.rbegin();
	if(it!=r.rend()) ++it;
	if(it!=r.rend()){
		Q = (*it)->Q;
		P = (*it)->P;
	}
	else{
		Q = INT_MAX;
		P = 1.0;
	}
}
#endif

vector<int>* GetPath(const Route& r){
#ifndef BUG1
	if(r.size()==0)return new vector<int>();
#else
	if(r.size()==0)return 0;
#endif
	vector<int>* path = new vector<int>(r.size()+1);
	Route::const_iterator it = r.begin();
	(*path)[0] = (*it)->channel->from;
	int i=1;
	while(it!=r.end()){
		(*path)[i++] = (*it)->channel->to;
		++it;
	}
	return path;
}

ostream& operator<<(ostream& os,const Route& r){
	vector<int>* path = GetPath(r);
	for(int i=0; i<path->size(); i++)
		os<<(*path)[i]<<" ";
	delete path;
	os<<endl<<"E="<<GetE(r);
	return os;
}

//**********************************************************

class Routers{
public:
	//Constructor:
	Routers(int N,int from=0,int to=1):channelsFrom(N+1),channelsTo(N+1),
			From(from),To(to),cicles(N+1),currRoute(),bestRoute(),distTo(N+1,MAX_DIST),
#ifdef CUTOFF
			numOfBad(N+1),
			cutoffTrigger(N+1),
			cutoff(false),
			cutoffMin(MAX_DIST),
#ifdef CACHE_WOULDBE_BAD_LINKS
			badInd(),//Indices of bad links in currRoute vector
#endif
#endif
#ifdef OPTIMIZE_EQUAL_LINKS
			minQ(INT_MAX),maxQ(),
			minP(1.0),maxP(0.0),
			minE((double)INT_MAX),maxE(0.0),
#endif
#ifdef SET_BAD_LINKS
			badRecalc(false),
#ifdef STATS
			nBadRecalc(0),
#endif
#endif
#ifdef STATS
			nLinks(0),nBad(0),nDistTo(-1),
#endif
			currE(-1),bestE(-1)
#ifndef USE_VECTOR_ROUTES
			{}
#else
			{
				currRoute.reserve(N);
				bestRoute.reserve(N);
#ifdef CACHE_WOULDBE_BAD_LINKS
				badInd.reserve(N);
#endif
			}
#endif

	void SetFromTo(int f,int t){From=f;To=t;}

	int size(){return channelsFrom.size();}

	//***********************************************************
	void addChannel(int f, int t, int q, double p){
		Channel* pch = new Channel(f,t,q,p,channelsFrom[f].size(),channelsTo[t].size());
		channelsFrom[f].push_back(pch);
		pch = new Channel(f,t,q,p,channelsFrom[f].size()-1,channelsTo[t].size());
		channelsTo[t].push_back(pch);
#ifdef OPTIMIZE_EQUAL_LINKS
		if(q<minQ)minQ=q;
		if(p<minP)minP=p;
		if(double e=(double)q*p<minE)minE=e;
		if(q>maxQ)maxQ=q;
		if(p>maxP)maxP=p;
		if(double e=(double)q*p>maxE)maxE=e;
#endif
#ifdef STATS
		nLinks++;
#endif
	}

	//***********************************************************
	void Init(timer* time=0){
		if(time!=0)
			cout<<"Calculating distances...";
		SetDistTo();
		if(time!=0){
			time->fix();
			cout<<"Done! (";
			time->print(cout);
			cout<<")."<<endl;
		}
#ifdef FIND_SHORTEST_ROUTE
		//if(time!=0)
		FindShortestRoute();
		if(time!=0){
			if(bestRoute.size()){
				cout<<endl<<"Shortest route: "<<bestRoute<<endl;
			}
		}
#ifdef SET_BAD_LINKS
#ifdef CUTOFF
		cutoffMin = MAX_DIST;
		cutoff = false;
#endif
		BadLinks();
		badRecalc = false;
#ifdef CUTOFF
		if(cutoff) SetDistTo(cutoffMin);
#endif
#endif
#endif
	}

	//***********************************************************
	//This can be excluded by modifying NextFirst() a little;
	void First(){
		Route& pr = currRoute;
		if(distTo[From]==MAX_DIST) return;//No routes from From to To
		int from = From;
		cicles[from] = 1;
#if FIND_BAD_LINKS>1
		int Q = INT_MAX;
		double P = 1.0;
#endif
		while(true){
#if FIND_BAD_LINKS>1
			if(pr.size()>0){
				Q = pr.back()->Q;
				P = pr.back()->P;
			}
#endif
			if(channelsFrom[from].size()==0) break;
			vector<Channel*>* pc = &channelsFrom[from];
#if FIND_BAD_LINKS<2
			int i=0;
			while(i<pc->size()&&(cicles[channelsFrom[from][i]->to]==1||distTo[(*pc)[i]->to]==MAX_DIST
					//||(*pc)[i]->isBad
					))i++;
#else
			for(int i=0;i<pc->size();i++){
				if(cicles[channelsFrom[from][i]->to]==1)continue;
				if(distTo[(*pc)[i]->to]==MAX_DIST)continue;
				currE=(double)(min(Q,(*pc)[i]->Q)*P*(*pc)[i]->P);
				if(LEQ(currE,bestE)){
					if(EQUIV(currE,double((*pc)[i]->Q)*(*pc)[i]->P)){
						(*pc)[i]->isBad = true;
						channelsTo[(*pc)[i]->to][(*pc)[i]->indT]->isBad = true;
#ifdef STATS
						nBad++;
#endif
#ifdef SHOW_BAD_LINKS
	 					cout<<"Bad link(F): "<<(*pc)[i]->from<<" "<<(*pc)[i]->to
	 							<<" currE="<<currE<<" "<<"linkE="<<double((*pc)[i]->Q)*(*pc)[i]->P<<endl;
#endif
#ifdef CUTOFF
						if(++numOfBad[from]==cutoffTrigger[from])
							cutoff = true;
							if(cutoffMin>distTo[from]-1)
								cutoffMin = distTo[from]-1;
#endif
					}
					continue;
				}
#ifdef CACHE_WOULDBE_BAD_LINKS
				if(EQUIV(currE,double((*pc)[i]->Q)*(*pc)[i]->P)){//See for potentially bad
					if(badInd.size()!=0){
						Channel* ch = pr[badInd[badInd.size()-1]]->channel;
						if(LESS(double((*pc)[i]->Q)*(*pc)[i]->P,double(ch->Q)*ch->P))
							badInd.clear();
					}
					//if(pr.size()!=0)
					badInd.push_back(pr.size()/*-1 this will be the NEXT link*/);
				}
#endif
				break;
			}
#endif
			//cout<<"currRoute: "<<pr<<endl;
			if(i==pc->size()) break;//this gives no guarantee that the first route (if not empty) ends in To
			append(pr,channelsFrom[from][i]);
			from = channelsFrom[from][i]->to;
			cicles[from] = 1;
			if(from==To) break;
			currE= double(pr.back()->P*pr.back()->Q);
			if(LEQ(currE,bestE))
				break;
		}
#if FIND_BAD_LINKS>1
#ifndef BUG4//Fixed 9.2.01
//First route may turn out to be the last one as well!
//E.g: links 1->2,2->1,1->3, with shortest route 1->3
		if(pr.size()>0){
#endif
#endif
			currE= double(pr.back()->P*pr.back()->Q);
#ifndef BUG0//Bug fix 31.1.01
		if(pr.back()->channel->to==To)//forgot to fix this bug here as well!
#endif
		if(LESS(bestE,currE)){
			bestE=currE;
#ifndef BUG3
			copy(bestRoute,pr);
#else
			bestRoute = pr;
#endif
#ifdef SET_BAD_LINKS
			//BadLinks();//Strangely highly slows down calculations!
#endif
		}
#if FIND_BAD_LINKS>1
#ifndef BUG4//Fixed 9.2.01
		}
#endif
#endif
		//cout<<pr<<endl;
#ifdef CACHE_WOULDBE_BAD_LINKS
		if(LEQ(currE,bestE))//Otherwise inadequate bad links may appear
	 	for(int i=0;i<badInd.size();i++){
	 			Channel* ch = pr[badInd[i]]->channel;
	 			ch->isBad = true;
	 			channelsTo[ch->to][ch->indT]->isBad = true;
#ifdef SHOW_BAD_LINKS
				cout<<"Bad link1: "<<ch->from<<" "<<ch->to<<endl;
#endif
#ifdef CUTOFF
	 			if(++numOfBad[ch->from]==cutoffTrigger[ch->from])
	 			cutoff = true;
				if(cutoffMin>distTo[ch->from]-1)
					cutoffMin = distTo[ch->from]-1;
#endif
#ifdef STATS
				nBad++;
#endif
	 	}
		else
			badInd.clear();//Clear for later use
#ifndef QUICK_MOUNT
		badInd.clear();//Clear for later use
#endif
#endif
		return;
	}

	//***********************************************************
	void NextFirst(){
		if(currRoute.size()==0) return;//No more routes from From to To
		Route& pr = currRoute;
		int from = pr.back()->channel->to;
		while(true){
			//if(pr.size()==0) return;//This can not (?) happen
	 		int Q = pr.back()->Q;
	 		double P = pr.back()->P;
	 		if(channelsFrom[from].size()==0) break;
	 		vector<Channel*>* pc = &channelsFrom[from];
#if FIND_BAD_LINKS<2
	 		int i=0;
	 		while(i<pc->size()&&(LEQ(double(min(Q,(*pc)[i]->Q))*P*(*pc)[i]->P,bestE)||
	 				cicles[channelsFrom[from][i]->to]==1||distTo[(*pc)[i]->to]==MAX_DIST
	 				||(*pc)[i]->isBad
	 				))i++;
#else
	 		for(int i=0;i<pc->size();i++){
	 			if(cicles[channelsFrom[from][i]->to]==1)continue;
	 			if(distTo[(*pc)[i]->to]==MAX_DIST)continue;
	 			if((*pc)[i]->isBad)continue;
	 			currE = double(min(Q,(*pc)[i]->Q))*P*(*pc)[i]->P;
	 			if(LEQ(currE,bestE)){
	 				if(EQUIV(currE,double((*pc)[i]->Q)*(*pc)[i]->P)){
	 					(*pc)[i]->isBad = true;
	 					channelsTo[(*pc)[i]->to][(*pc)[i]->indT]->isBad = true;
#ifdef STATS
				nBad++;
#endif
#ifdef SHOW_BAD_LINKS
	 					cout<<"Bad link(F): "<<(*pc)[i]->from<<" "<<(*pc)[i]->to
	 							<<" currE="<<currE<<" "<<"linkE="<<double((*pc)[i]->Q)*(*pc)[i]->P<<endl;
#endif
#ifdef CUTOFF
	 					if(++numOfBad[from]==cutoffTrigger[from])
	 						cutoff = true;
	 					if(cutoffMin>distTo[from]-1)
	 						cutoffMin = distTo[from]-1;
#endif
	 				}
	 				continue;
	 			}
#ifdef CACHE_WOULDBE_BAD_LINKS
	 			if(EQUIV(currE,double((*pc)[i]->Q*(*pc)[i]->P))){//See for potentially bad
	 				if(badInd.size()!=0){
	 					Channel* ch = pr[badInd[badInd.size()-1]]->channel;
	 					if(LESS(double((*pc)[i]->Q)*(*pc)[i]->P,double(ch->Q)*ch->P))
	 						badInd.clear();
	 				}
	 				badInd.push_back(pr.size()/*-1 this will be the NEXT link*/);
	 			}
#endif
				break;
			}
#endif
			if(i==pc->size()) break;
			append(pr,channelsFrom[from][i]);
			from = channelsFrom[from][i]->to;
			cicles[from] = 1;
			if(from==To) break;
			currE= double(pr.back()->P*pr.back()->Q);
			if(LEQ(currE,bestE))
				break;
		}
		currE= double(pr.back()->P*pr.back()->Q);
		if(pr.back()->channel->to==To)
		if(LESS(bestE,currE)){
			bestE=currE;
#ifndef BUG3
			copy(bestRoute,pr);
#else
			bestRoute = pr;
			//cout<<"Best route: "<<bestRoute<<endl;
#endif
#ifdef SET_BAD_LINKS
			//BadLinks();//Strangely highly slows down calculations!
#endif
		}
#ifdef CHOOSE_SHORTER_BEST//This add on should not slow down the algorithm
		else
		if(EQUIV(currE,bestE)&&pr.size()<bestRoute.size())
#ifndef BUG3
			copy(bestRoute,pr);
#else
			bestRoute = pr;
#endif
#endif
#ifdef CACHE_WOULDBE_BAD_LINKS
		if(LEQ(currE,bestE))//Otherwise inadequate bad links may appear
	 	for(int i=0;i<badInd.size();i++){
	 			Channel* ch = pr[badInd[i]]->channel;
	 			ch->isBad = true;
	 			channelsTo[ch->to][ch->indT]->isBad = true;
#ifdef SHOW_BAD_LINKS
				cout<<"Bad link1: "<<ch->from<<" "<<ch->to<<endl;
#endif
#ifdef CUTOFF
	 			if(++numOfBad[ch->from]==cutoffTrigger[ch->from])
	 			cutoff = true;
				if(cutoffMin>distTo[ch->from]-1)
					cutoffMin = distTo[ch->from]-1;
#endif
#ifdef STATS
				nBad++;
#endif
	 	}
		else
			badInd.clear();//Clear for later use
#ifndef QUICK_MOUNT
		badInd.clear();//Clear for later use
#endif
#endif
		//cout<<"Curr route: "<<currRoute<<endl;
		return;
	}

	//***********************************************************
	void Next(){
		Route& pr = currRoute;
#ifdef QUICK_MOUNT

#ifdef CUTOFF
			if(cutoff==true){
				SetDistTo(cutoffMin);
				Route::iterator it = pr.begin();
				while(it!=pr.end()&&distTo[(*it)->channel->from]!=MAX_DIST
						&&!(*it)->channel->isBad)++it;
				//pr.erase(it,pr.end());//memory leaks
				clear(pr,it);
			}
#endif
#if 0//Exclude this for now
#if defined SET_BAD_LINKS&&defined CUTOFF
			else
#endif
#ifdef SET_BAD_LINKS
			if(badRecalc==true){
				Route::iterator it = pr.begin();
				while(it!=pr.end()&&distTo[(*it)->channel->from]!=MAX_DIST
						&&!(*it)->channel->isBad)++it;
				//pr.erase(it,pr.end());//memory leaks
				clear(pr,it);
				badRecalc = false;
			}
#endif
#endif //0
#if defined CACHE_WOULDBE_BAD_LINKS&&(defined CUTOFF||defined SET_BAD_LINKS)
			else
#endif
#ifdef CACHE_WOULDBE_BAD_LINKS
			if(badInd.size()!=0){
				//pr.erase(pr.begin()+badInd[0]+1,pr.end());//memory leaks
				clear(pr,pr.begin()+badInd[0]+1);
			}
			badInd.clear();
#endif

#if FIND_BAD_LINKS>1
#if undef CUTOFF&&undef SET_BAD_LINKS&&undef CACHE_WOULDBE_BAD_LINKS
			Route::iterator it = pr.begin();
			while(it!=pr.end()&&distTo[(*it)->channel->from]!=MAX_DIST
				&&!(*it)->channel->isBad)++it;
			//pr.erase(it,pr.end());//memory leaks
			clear(pr,it);
#endif
#endif

#endif//QUICK_MOUNT
#ifdef CUTOFF
			cutoff = false;
			cutoffMin = MAX_DIST;
#endif
		while(true){
			//cout<<endl<<pr<<endl;
			if(pr.size()==0/*&&bestE!=-1*/) return;//No more routes from From to To
			int from = pr.back()->channel->from;
			int i = pr.back()->channel->indF+1;
			int Q;
			double P;
			GetQPprev(pr,Q,P);
#if FIND_BAD_LINKS!=0&&FIND_BAD_LINKS!=2//Check for possible cutoff and Bad Links
//when FIND_BAD_LINKS>1 then (potentially) bad links are searched
//in NextFirst() as well
			Channel* ch = channelsFrom[from][pr.back()->channel->indF];//current last link
			double e = (double)(ch->Q*ch->P);
			if(EQUIV(e,currE)||LEQ(e,bestE)){//effectivity of route equals that of its last link
				ch->isBad = true;
				channelsTo[ch->to][ch->indT]->isBad = true;
				numOfBad[from]++;
#ifdef SHOW_BAD_LINKS
				cout<<"Bad link(N0): "<<ch->from<<" "<<ch->to
						<<" currE="<<currE<<" "<<"linkE="<<double(ch->Q)*ch->P<<endl;
#endif
#ifdef STAT
				nBad++;
#endif
#ifdef CUTOFF
				if(numOfBad[from]==cutoffTrigger[from])
					cutoff = true;
				if(cutoffMin>distTo[from]-1)
					cutoffMin = distTo[from]-1;
#endif
			}
#endif
			vector<Channel*>* pc = &channelsFrom[from];
#if FIND_BAD_LINKS!=2
			while(i<pc->size()&&(LEQ(double(min(Q,(*pc)[i]->Q))*P*(*pc)[i]->P,bestE)||
					cicles[channelsFrom[from][i]->to]==1||distTo[(*pc)[i]->to]==MAX_DIST
					||(*pc)[i]->isBad
					))i++;
#else
			for(/*i=0Bug!!!*/;i<pc->size();i++){
				if(cicles[channelsFrom[from][i]->to]==1)continue;
				if(distTo[(*pc)[i]->to]==MAX_DIST)continue;
				if((*pc)[i]->isBad)continue;
				double//Make it local here
				currE=(double)(min(Q,(*pc)[i]->Q)*P*(*pc)[i]->P);
				if(LEQ(currE,bestE)){
					if(EQUIV(currE,double((*pc)[i]->Q*(*pc)[i]->P))){
						(*pc)[i]->isBad = true;
						channelsTo[(*pc)[i]->to][(*pc)[i]->indT]->isBad = true;
#ifdef STATS
						nBad++;
#endif
#ifdef SHOW_BAD_LINKS
						cout<<"Bad link(N): "<<(*pc)[i]->from<<" "<<(*pc)[i]->to
								<<" currE="<<currE<<" "<<"linkE="<<double((*pc)[i]->Q)*(*pc)[i]->P<<endl;
#endif
#ifdef CUTOFF
						if(++numOfBad[from]==cutoffTrigger[from])
							cutoff = true;
						if(cutoffMin>distTo[from]-1)
							cutoffMin = distTo[from]-1;
#endif
					}
					continue;
				}
#ifdef CACHE_WOULDBE_BAD_LINKS
				if(EQUIV(currE,double((*pc)[i]->Q*(*pc)[i]->P))){//See for potentially bad
					if(badInd.size()!=0){
						Channel* ch = pr[badInd[badInd.size()-1]]->channel;
						if(LESS(double((*pc)[i]->Q)*(*pc)[i]->P,double(ch->Q)*ch->P))
							badInd.clear();
					}
					badInd.push_back(pr.size()-1/*Here -1 is OK*/);
				}
#endif
				break;
			}
#endif
			cicles[pr.back()->channel->to] = 0;
#ifndef BUG3//memory leaks, eating heap
			popAndClean(pr);
#else
			pr.pop_back();
#endif
			if(i==pc->size()) continue;
#if FIND_BAD_LINKS==1
#ifdef CUTOFF
			if(cutoff) SetDistTo(cutoffMin);
#endif
#endif
			assert(channelsFrom[from][i]->from<2000);
			append(pr,channelsFrom[from][i]);
			cicles[channelsFrom[from][i]->to] = 1;
			NextFirst();
			break;
		}
		//cout<<endl<<pr;
		return;
	}

	//***********************************************************
	Route& GetBestRoute(){
#ifdef OPTIMIZE_EQUAL_LINKS
#ifndef APPROX_BEST
		if(minQ==maxQ&&minP==maxP) return bestRoute;//Shortest route is best
#else
		if(equiv(pow(double(minQ)*minP,bestRoute.size()),pow(double(maxQ)*maxP,bestRoute.size())))return bestRoute;//Shortest route is best
#endif
#endif
		First();
		while(currRoute.size()>0)
			Next();
		return bestRoute;
	}

	//***********************************************************
	//This is between O(N*N) and O(N*N*N)
	void SetDistTo(int i=0){
		//if(i<0)i=0;
		distTo[To] = 0;
		vector<vector<int> > distTo2(size()-1);
		distTo2[0].push_back(To);
		int added = 1;
#ifdef OPTIMIZE_SET_DIST
		for(int j=0; j<size(); j++)
#ifndef CUTOFF
			if(j!=To&&channelsFrom[j].size()==0)added++;//The point is inaccessible from To
#else
			if(j!=To)
			if(channelsFrom[j].size()==numOfBad[j])
				{added++;distTo[j]=MAX_DIST;}//The point is inaccessible from To
			else
			if(distTo[j]>i)distTo[j]=MAX_DIST;//These are to be recalculated
			else
			if(distTo[j]==i){distTo2[i].push_back(j);added++;}
			else added++;
#endif
#endif
		for(; i<size()-1; i++){
			bool addedNew = false;
			for(int j=0; j<distTo2[i].size(); j++){
				vector<Channel*>* chann = &channelsTo[distTo2[i][j]];
				for(int k=0; k<chann->size(); k++){
					int from = (*chann)[k]->from;
					if(distTo[from]==MAX_DIST&&!(*chann)[k]->isBad){
						distTo2[i+1].push_back(from);
						distTo[from] = i+1;
						addedNew = true;
						added++;
					}
#ifdef CUTOFF
					if(distTo[from]==i+1)cutoffTrigger[from]++;
#endif
				}
			}
			if(addedNew==false) break;//remaining routers are inaccessible
			if(added==size()) break;
		}
#ifdef SHOW_DIST
		cout<<endl<<"i="<<cutoffMin<<" distTo: ";
		for(i=1;i<size()-1;i++)
			cout<<distTo[i]<<" ";
		cout<<endl;
		//cout<<"Dist13="<<distTo[13]<<endl;//OK
#endif
#ifdef STATS
		nDistTo++;
#endif
#ifdef CUTOFF
		cutoff = false;
		cutoffMin = MAX_DIST;
#endif
	}

#ifdef FIND_SHORTEST_ROUTE
	void FindShortestRoute(){//and set it as the first bestRoute
		if(distTo[From]==MAX_DIST)
			return;
		int from = From;
		for(int i=distTo[From]-1;i>=0;i--){
			int j = 0;
			while(distTo[channelsFrom[from][j]->to]>i)j++;
			append(bestRoute,channelsFrom[from][j]);
			from = channelsFrom[from][j]->to;
		}
		bestE = GetE(bestRoute);
	}
#endif

#ifdef QUICK_MOUNT
	//***********************************************************
	void clear(Route&r, Route::iterator it){
		int i=0;
		while(it!=r.end()){
			cicles[(*it)->channel->to] = 0;//Bug fix;
			delete *it;
			++it;
			i++;
		}
		r.resize(r.size()-i);
	}
#endif

#ifdef SET_BAD_LINKS
	//***********************************************************
	void BadLinks(){
		if(bestE<=0) return;
#ifdef CUTOFF
		//cutoffMin = MAX_DIST;
		//cutoff = false;
#endif
		for(int i=0; i<size(); i++)
		for(int j=0; j<channelsFrom[i].size(); j++){
			if(i==To) continue;
			Channel* ch = channelsFrom[i][j];
			if(LEQ(double(ch->Q)*ch->P,bestE)&&!ch->isBad){
				ch->isBad = true;
				channelsTo[ch->to][ch->indT]->isBad = true;
				numOfBad[i]++;
				badRecalc = true;
#ifdef SHOW_BAD_LINKS
				cout<<"Bad link(N0): "<<ch->from<<" "<<ch->to
						<<" currE="<<currE<<" "<<"linkE="<<double(ch->Q)*ch->P<<endl;
#endif
#ifdef CUTOFF
				if(numOfBad[i]==cutoffTrigger[i])
					cutoff = true;
				if(cutoffMin>distTo[i]-1)
					cutoffMin = distTo[i]-1;
#endif
#ifdef STATS
				nBad++;
#endif
			}
		}
#ifdef STATS
		nBadRecalc++;
#endif
#ifdef CUTOFF
		//if(cutoff) SetDistTo(cutoffMin);
#endif
	}
#endif

protected:
	Route currRoute,bestRoute;
	double currE, bestE;
	int From, To;//source and target routers
	vector<vector<Channel*> > channelsFrom;
	vector<vector<Channel*> > channelsTo;
	vector<int> cicles;//indicator of routes with loops
	//vector<int> distFrom;//minimal distances from From router//not needed
	vector<int> distTo;//minimal distances from To router
#ifdef CUTOFF
	vector<int> numOfBad;//current number of bad links from a given router
	vector<int> cutoffTrigger;//number of links from a given node with smaller distTo
	bool cutoff;//Flag showing that it is time to recalculate distances;
	int cutoffMin;//minimal cutoff distance To
#ifdef CACHE_WOULDBE_BAD_LINKS
	vector<int> badInd;//Indices of bad links in currRoute vector
#endif
#endif
#ifdef OPTIMIZE_EQUAL_LINKS
	//A bit more data are saved than are used now
	//(a simple bool flag is sufficient);
	//but may be useful in future optimizations
	int minQ,maxQ;
	double minP,maxP,minE,maxE;
#endif
#ifdef SET_BAD_LINKS
	bool badRecalc;
#endif
#ifdef STATS
public:
	int nLinks,nBad,nDistTo;
#ifdef SET_BAD_LINKS
	int nBadRecalc;
#endif
#endif
};

int main(int argc, char **argv)
{
	enum TESTS{NOTEST,TEST};

	//Parameters changeable from command line (for debug tests):
	TESTS test = NOTEST;//Default is no test run
	int clusters = 1;
	int N0 = 300;//cluster size
	int Br = 0;//Neighbour clusters are connected with Br bridge routers
	int Q0 = 64000;
	int deltaQ = 1;
	double P0 = 1.0;
	double deltaP = 0.0;
	int QBridge = 32000;
	double PBridge = 0.1;
	double deltaPBridge = 0.1;
	int sign = 1;
	bool verbose = false;
	//as well as global variable epsilon
	//(setting  precision of best effectivity search)

	int N,M;
	ifstream is;

	//Command line options -- for debug tests
	for (int OptNum = 1; OptNum < argc; OptNum++)
	{
		string str = strupr(argv[OptNum]);
		string i_str = str.substr(1);
		char ch;

		switch(str[0]){
	 		case 'T' :
				test = TEST;
	 			break;
	 		case 'V' :
				verbose = true;
	 			break;
	 		case 'E' :
				epsilon = atof(i_str.c_str());
	 			break;
	 		case 'C' :
				clusters = atoi(i_str.c_str());
	 			goto DEFAULT;
	 		case 'N' :
				N0 = atoi(i_str.c_str());
	 			goto DEFAULT;
	 		case 'B' :
				Br = atoi(i_str.c_str());
	 			goto DEFAULT;
	 		case 'S' :
				sign = atoi(i_str.c_str());
	 			goto DEFAULT;
	 		case 'Q' :
				ch = i_str[0];
				if(isdigit(ch))
					Q0 = atoi(i_str.c_str());
				else
				if(ch=='D'){
					i_str = str.substr(2);
					deltaQ = atoi(i_str.c_str());
				}
	 			goto DEFAULT;
	 		case 'P' :
				ch = i_str[0];
				if(isdigit(ch))
					P0 = atof(i_str.c_str());
				else
				if(ch=='D'){
					i_str = str.substr(2);
					deltaP = atof(i_str.c_str());
				}
				else
				if(ch=='B'){
					ch = i_str[1];
					if(isdigit(ch)){
						i_str = str.substr(2);
						PBridge = atof(i_str.c_str());
					}
					else
					if(ch=='D'){
						i_str = str.substr(3);
						deltaPBridge = atof(i_str.c_str());
					}
				}
	 			//goto DEFAULT;
			DEFAULT:
			default:
				test = TEST;
	 	}
	}

if(test==TEST)
	N = clusters?N0*clusters+Br*(clusters-1):3;
else{
	is.open("NETWORK.INP");
	if(!is){
		cout<<"Can not open NETWORK.INP!"<<endl;
		exit(1);
	}
	is>>N>>M;
}
	Routers routers(N+1);

if(test==TEST){
//Create Routers consisting of clusters clusters
//(complete oriented graphs of size N0),
//neighbor clusters connected with Br "Bridges" (potential
//bad links (in fact they are to be called "good", because may reduce
//exponential search to polinomial one :)).
//Varying other parameters, one can create arbitrary complex
//(or simple) configurations for testing.
	int i,j,cl;
	for(cl=1;cl<=clusters;cl++){
		int start = (cl-1)*(N0+Br)+1;
		for(i=start;i<start+N0;i++)
		for(j=start;j<start+N0;j++)
		if(i!=j)
			routers.addChannel(i,j,max(1,Q0-deltaQ*(j+sign*i)),min(1.0,max(0.0,P0-deltaP*double(j+sign*i))));
		if(cl<clusters)
		for(i=1;i<=Br;i++){
			routers.addChannel(start+N0-1,start+N0-1+i,QBridge,max(0.0,min(1.0,PBridge+deltaPBridge*(double)i)));
			routers.addChannel(start+N0-1+i,start+N0+Br,QBridge,1);
		}
	}
	if(clusters==0){//To catch BUG4
		routers.addChannel(1,2,1,1.0);
		routers.addChannel(2,1,1,1.0);
		routers.addChannel(1,3,1,1.0);
		routers.SetFromTo(1,3);
	}
	else
		routers.SetFromTo(1,N0*clusters+Br*(clusters-1));
//Some test results:
//This runs 1 min 20 sec at N0==300 clusters==2 and Br==4 (604 nodes in general);
//and it takes <=1 sec if N0=100 and Br==10 (210 nodes)
//but for Br=11 I failed to wait to the end of search!
//If clusters==1 and sign==+1, then for N0==500 search takes 45 sec (with all optimizations on)
//and 17 swc for N0==300.
//With SET_BAD_LINKS undefined this takes exponential time even for N0==17!
//There are no such problems for sign==-1.
}
else{
	while(!is.eof()){
		int from,to;
		int Q;
		double P;
		for(int i=0;i<M;i++){
			is>>from>>to>>Q>>P;
			routers.addChannel(from,to,Q,P);
		}
		is>>from>>to;
		routers.SetFromTo(from,to);
	}
	is.close();
}

	ofstream os("NETWORK.OUT");
	if(!os){
		cout<<"Can not open Catalog.out!"<<endl;
		exit(1);
	}

	if(test==TEST){
		os<<"Routers config:"<<endl
			<<"Nodes="<<N
#ifdef STATS
			<<" Links="<<routers.nLinks
#endif
			<<" Clusters="<<clusters
			<<" cluster size="<<N0<<" bridges="<<Br<<endl
			<<"Q="<<Q0<<" Delta Q="<<deltaQ
			<<" P="<<P0<<" Delta P="<<deltaP<<endl
			<<"QBridge="<<QBridge<<" PBridge="<<PBridge
			<<" Delta PBridge="<<deltaPBridge<<endl
			<<"Sign="<<sign<<" epsilon="<<epsilon<<endl;
	}
	timer time;
	time_t curr_time = 0;
	if(verbose){
		cout<<"Searching for the best route. Wait..."<<endl;
		routers.Init(&time);
	}
	else
		routers.Init();
	Route best = routers.GetBestRoute();
	time.fix();

	if(verbose){
		cout<<endl<<"Best route: ";
		cout<<best<<endl<<endl<<"Done! (";
		time.print(cout);
		cout<<")."<<endl;
		//cout<<endl;
	}

	vector<int>* path = GetPath(best);

if(test==TEST){
	if(best.size()==0)
		os<<"No routes found";
	else{
		os<<endl<<"Best route: "
			<<setiosflags(ios::fixed)<<endl;
 		for(int i=0; i<path->size(); i++){
			os<<(*path)[i];
			if((i+1)%16!=0)
				os<<" ";
			else
				os<<endl;
		}
		os<<endl<<endl<<"Eff.="<<GetE(best)<<" "<<" Routes="<<best.size()+1
#ifdef STATS
			<<endl<<endl<<"Bad Links: "<<routers.nBad<<" DistTo recalc.: "<<routers.nDistTo
#ifdef SET_BAD_LINKS
			<<" Bad Links recalc.: "<<routers.nBadRecalc
#endif
#endif
			<<endl<<endl<<"Found in ";
		time.print(os);
		os<<"."<<endl;
	}
}
else{
	if(best.size()==0){
		os<<0<<" "<<0<<endl<<endl;
		return 0;
	}
	os<<setiosflags(ios::fixed)<<GetE(best)<<" "<<best.size()+1<<endl;
	vector<int>* path = GetPath(best);
 	for(int i=0; i<path->size(); i++){
		os<<(*path)[i];
		if(i<path->size()-1)
			os<<" ";
	}
}
	delete path;
	os.close();
	return 0;
}